#ifndef WTLIB_H
#define WTLIB_H

#include <complex.h>

/*****************************************************************************/
/* create_wf: Creates a new wavelet family of functions.                     */
/*****************************************************************************/
/*  WFid (Output): Wavelet family structure identifier.                      */
/*  type (Input):  Mother wavelet function type.                             */
/*    -1: Complex Morlet Wavelet.                                            */
/*    -2: Zero mean Complex Morlet Wavelet.                                  */
/*  J (Input):  Number of octave.                                            */
/*  V (Input):  Number of voices per octave.                                 */
/*  N (Input):  Maximum wavelet length.                                      */
/*  s0 (Input): Minimum scale.                                               */
/*  b0 (Input): Sampling at the minimum scale (just in the discrete case).   */
/*  convtype (Input): type of convolution performed.                         */
/*    0: circular. (Other to come)                                           */
/*  w0 (Input): central frequency of the mother wavelet.                     */
/*  continuous: uniform sampling (1) vs dyadic sampling (0).                 */
/*****************************************************************************/
int create_wf_ (unsigned int *WFid, int *type, unsigned int *J, 
                unsigned int *V, unsigned int *N, double *s0, double *b0, 
                int *convtype, double *w0, int *continuous);

/*****************************************************************************/
/* destroy_wf: Destroys a wavelet family of functions.                       */
/*****************************************************************************/
/*  WFid (Output): Wavelet family structure identifier.                      */
/*****************************************************************************/
int destroy_wf_ (unsigned int *WFid);

/*****************************************************************************/
/* fwts: Wavelet transform of a single-precision real sequence.              */
/*****************************************************************************/
/* WTx (Output): Wavelet-transformed sequences (double-precision complex     */
/*               matrix).                                                    */
/* x (Input): Single-precision real sequence to transform.                   */
/* N (Input): Length of the sequence to transform.                           */
/* Nmax (Input): Length of the time dimension of the WTx matrix (min = N).   */ 
/* WFid (Input): Wavelet family structure identifier.                        */
/* STlike (Input): Add the S-transform make up to the Morlet wavelet trans.  */
/*                 ST like (1) / WT like (0)                                 */
/* dt (Input): Sampling period (only used when STlike = 1.                   */
/*****************************************************************************/
int fwts_ (double complex *WTx, float *x, unsigned int *N, unsigned int *Nmax, 
           unsigned int *WFid, int STlike, double *dt);

/*****************************************************************************/
/* ifwts: Inverse wavelet transform giving a single-precision real sequence. */
/*****************************************************************************/
/* xrec (Output): Synthetisized single-precision real sequence.              */
/* WTx (Iput): Wavelet-transformed sequences (double-precision complex       */
/*               matrix).                                                    */
/* N (Input): Length of the sequence to transform.                           */
/* Nmax (Input): Length of the time dimension of the WTx matrix (min = N).   */ 
/* WFid (Input): Wavelet family structure identifier.                        */
/* STlike (Input): Add the S-transform make up to the Morlet wavelet trans.  */
/*                 ST like (1) / WT like (0)                                 */
/* dt (Input): Sampling period (only used when STlike = 1.                   */
/*****************************************************************************/
int ifwts_ (float *xrec, double complex *WTx, unsigned int *Ns, unsigned int *N, 
            unsigned int *Nmax, unsigned int *WFid, int STlike, double *dt);	

/*****************************************************************************/
/* fwt: Wavelet transform of a double-precision real sequence.               */
/*****************************************************************************/
/* Idem to fwts but being x a double-precision real sequence.                */
/*****************************************************************************/
int fwt_ (double complex *WTx, double *x, unsigned int *N, unsigned int *Nmax, 
          unsigned int *WFid, int STlike, double *dt);

/*****************************************************************************/
/* fwt:  Inverse wavelet transform giving a single-precision real sequence.  */
/*****************************************************************************/
/* Idem to fwt but being xreac a double-precision real sequence.             */
/*****************************************************************************/

int ifwt_ (double *xrec, double complex *WTx, unsigned int *Ns, unsigned int *N, 
           unsigned int *Nmax, unsigned int *WFid, int STlike, double *dt);

#endif
